#include <linux/init.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/ipv6.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/netlink.h>

#include "pla_crypto.h"
#include "pla_crypto_impl.h"

static struct pla_crypto_impl *crypto_impl;

/*
 *  Returns: 1: success
 *           0: command not supported.
 *          -1: error
 */
int pla_crypto_netlink(struct nlmsghdr *nlh)
{
	if (crypto_impl->crypto_netlink)
		return crypto_impl->crypto_netlink(nlh);

	return 0;
}


int pla_crypto_call(
	void (*callback) (void *call_state),
	void *state,
	struct pla_crypto_task *task)
{
	int err = 0;

	if (!crypto_impl) {
		err = -EINVAL;
		goto out;
	}

	err = crypto_impl->call_crypto(callback, state, task);

out:
	return err;
}


int pla_crypto_init(int impl)
{
	int err = 0;

	crypto_impl = pla_crypto_get_impl(impl);

	if (!crypto_impl) {
		err = -EINVAL;
		goto out;
	}

	if (crypto_impl->crypto_impl_init)
		err = crypto_impl->crypto_impl_init();
	if (err < 0)
		goto out;

out:
	return 0;
}

void pla_crypto_uninit(void)
{
	if (crypto_impl->crypto_impl_uninit)
		crypto_impl->crypto_impl_uninit();

	return;
}
