#ifndef _PLA_CRYPTO_H
#define _PLA_CRYPTO_H

#include <linux/types.h>
#include <linux/netlink.h>

#include "../src/configs.h"

int pla_crypto_init(int impl);
void pla_crypto_uninit(void);

#define CRYPTO_TASK_TYPE_SIGN   1
#define CRYPTO_TASK_TYPE_VERIFY 2

#define CRYPTO_TASK_TYPE_RESPONSE 4

#define CRYPTO_TASK_CODE_OK     1
#define CRYPTO_TASK_CODE_NOT_OK 2

struct crypto_sign_request
{
	__u8 hash[PLA_HASH_DIGEST_LENGTH];
	__u8 cert[PLA_CERT_LEN];
	__u8 privkey[PLA_PRIVKEY_LEN];
};

struct crypto_sign_response
{
	__u8 signature[PLA_SIG_LEN];
};

struct crypto_verify_request
{
	__u8 hash[PLA_HASH_DIGEST_LENGTH];
	__u8 cert[PLA_CERT_LEN];
	__u8 ttp_pubkey[PLA_PUBKEY_LEN];
	__u8 signature[PLA_SIG_LEN];
	__u8 id_hash[PLA_HASH_DIGEST_LENGTH];
};

struct crypto_verify_response
{
	__u8 result;
};

struct pla_crypto_task
{
	char type;
	char code;

	union {
		struct crypto_sign_request sreq;
		struct crypto_verify_request vreq;

		struct crypto_sign_response sresp;
		struct crypto_verify_response vresp;

		char raw_data[0];   /* Space allocated by other members */
	} td;

	unsigned data_len;          /* Length of data in the buffer */
};

int pla_crypto_call(
	void (*callback) (void *call_state),
	void *state,
	struct pla_crypto_task *task);

int pla_crypto_netlink(struct nlmsghdr *nlh);

#endif  // _PLA_CRYPTO_H
