#include <linux/module.h>
#include <linux/ipv6.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/version.h>

#include <net/ip.h>
#include <net/ipv6.h>
#include <net/tcp.h>
#include <net/icmp.h>
#include <linux/ip.h>
#include <linux/ipv6.h>

#include <asm/semaphore.h>
#include <asm/atomic.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>

#include <net/transp_v6.h>

#include "pla_header.h"
#include "configs.h"

static struct hbh_and_pla_hdr hbh_and_pla = 
{
	{ 0, HLEN/8-1 },              /* Hop-by-hop header */
	{ OPTION_PLA, HLEN - 4},      /* PLA option header */

};

/*
 * Return pointer to the beginning of the PLA payload.
 * Return NULL if not found.
 */
struct pla_hdr* pla_find_pla_header(struct sk_buff *skb)
{
	struct ipv6_hopopt_hdr *hop_hdr;

	char *raw_hbh;
	int offset, hbh_len;

	struct ipv6hdr *ip6hdr;
	
	ip6hdr = skb->nh.ipv6h;
	if (ip6hdr->nexthdr != NEXTHDR_HOP)
		return NULL;

	/* Hop-by-hop header is present. */
	hop_hdr = (struct ipv6_hopopt_hdr*) (ip6hdr+1);
	hbh_len = ipv6_optlen(hop_hdr);
	raw_hbh = (char *) hop_hdr;

	offset = sizeof(struct ipv6_hopopt_hdr);

	while(offset < hbh_len) {
		switch(raw_hbh[offset]) {
		case 0:
			offset++;
			break;
		case OPTION_PLA:
			return (struct pla_hdr*) &raw_hbh[offset+2];
			break;
		default:  /* Including case 1: */
			offset += raw_hbh[offset+1] + 2;
			break;
		}
	}

	return NULL;
}

struct ipv6_txoptions* pla_alloc_empty_options(struct sock *sk)
{
	struct ipv6_txoptions *opt;

	opt = sock_kmalloc(sk, sizeof(*opt), GFP_ATOMIC);
	if (!opt)
		return NULL;

	memset(opt, 0, sizeof(*opt));
	opt->hopopt = NULL;
	opt->dst0opt = NULL;
	opt->srcrt = NULL;
	opt->dst1opt = NULL;

	opt->tot_len = sizeof(*opt);

	return opt;
}

/*
 * Yes, this is code duplication from standard IPv6 stack. It cannot be avoided
 * as the standard code reads the options from user space.
 */ 
static int pla_ipv6_renew_option(void *ohdr, struct ipv6_opt_hdr *newopt, int newoptlen,
                             int inherit, struct ipv6_opt_hdr **hdr, char **p)
{
        if (inherit) {
                if (ohdr) {
                        memcpy(*p, ohdr, ipv6_optlen((struct ipv6_opt_hdr *)ohdr));
                        *hdr = (struct ipv6_opt_hdr *)*p;
                        *p += CMSG_ALIGN(ipv6_optlen(*(struct ipv6_opt_hdr **)hdr));
                }
        } else {
                if (newopt) {
			memcpy(*p, newopt, newoptlen);
                        *hdr = (struct ipv6_opt_hdr *)*p;
                        if (ipv6_optlen(*(struct ipv6_opt_hdr **)hdr) > newoptlen)
                                return -EINVAL;
                        *p += CMSG_ALIGN(newoptlen);
                }
        }
        return 0;
}

struct ipv6_txoptions *pla_ipv6_renew_options(struct sock *sk, struct ipv6_txoptions *opt,
					      int newtype, struct ipv6_opt_hdr *newopt, int newoptlen)
{
        int tot_len = 0;
        char *p;
        struct ipv6_txoptions *opt2;
        int err;

        if (newtype != IPV6_HOPOPTS && opt->hopopt)
                tot_len += CMSG_ALIGN(ipv6_optlen(opt->hopopt));
        if (newtype != IPV6_RTHDRDSTOPTS && opt->dst0opt)
                tot_len += CMSG_ALIGN(ipv6_optlen(opt->dst0opt));
        if (newtype != IPV6_RTHDR && opt->srcrt)
                tot_len += CMSG_ALIGN(ipv6_optlen(opt->srcrt));
        if (newtype != IPV6_DSTOPTS && opt->dst1opt)
                tot_len += CMSG_ALIGN(ipv6_optlen(opt->dst1opt));
        if (newopt && newoptlen)
                tot_len += CMSG_ALIGN(newoptlen);

        if (!tot_len)
                return NULL;

        tot_len += sizeof(*opt2);
        opt2 = sock_kmalloc(sk, tot_len, GFP_ATOMIC);
        if (!opt2)
                return ERR_PTR(-ENOBUFS);

        memset(opt2, 0, tot_len);

        opt2->tot_len = tot_len;
        p = (char *)(opt2 + 1);

        err = pla_ipv6_renew_option(opt->hopopt, newopt, newoptlen,
				    newtype != IPV6_HOPOPTS,
				    &opt2->hopopt, &p);
        if (err)
                goto out;

        err = pla_ipv6_renew_option(opt->dst0opt, newopt, newoptlen,
                                newtype != IPV6_RTHDRDSTOPTS,
                                &opt2->dst0opt, &p);
        if (err)
                goto out;

        err = pla_ipv6_renew_option(opt->srcrt, newopt, newoptlen,
                                newtype != IPV6_RTHDR,
                                (struct ipv6_opt_hdr **)opt2->srcrt, &p);
        if (err)
                goto out;

        err = pla_ipv6_renew_option(opt->dst1opt, newopt, newoptlen,
                                newtype != IPV6_DSTOPTS,
                                &opt2->dst1opt, &p);
        if (err)
                goto out;

        opt2->opt_nflen = (opt2->hopopt ? ipv6_optlen(opt2->hopopt) : 0) +
                          (opt2->dst0opt ? ipv6_optlen(opt2->dst0opt) : 0) +
                          (opt2->srcrt ? ipv6_optlen(opt2->srcrt) : 0);
        opt2->opt_flen = (opt2->dst1opt ? ipv6_optlen(opt2->dst1opt) : 0);

        return opt2;
out:
        sock_kfree_s(sk, opt2, opt2->tot_len);
        return ERR_PTR(err);
}


/* 
 * Add PLA hop-by-hop options into SOCK_STREAM sockets options. 
 */
int pla_add_pla_option(struct sk_buff *skb)
{
	struct tcp_sock *tp;

	struct ipv6_pinfo *np;
	struct ipv6_txoptions *opt = NULL;
	struct sock *sk = skb->sk;
	int retval = 0;

	np = inet6_sk(skb->sk);
	if (!np) 
		goto out;

	tp = tcp_sk(sk);
	if (!tp) 
		goto out;

	if (!np->opt) {
		np->opt = pla_alloc_empty_options(sk);
		if (!np->opt)
			goto out;
	}

	opt = pla_ipv6_renew_options(sk, np->opt, IPV6_HOPOPTS, &hbh_and_pla.hop_hdr, ipv6_optlen(&hbh_and_pla.hop_hdr) );
	if (IS_ERR(opt)) {
		printk(KERN_ALERT "Failed renew options\n");
		opt = NULL;
		goto out;
	}

	if ( !((1 << sk->sk_state) & (TCPF_LISTEN | TCPF_CLOSE)) && inet_sk(sk)->daddr != LOOPBACK4_IPV6)
		SYNCHRONIZE_MSS(sk,tp,opt);

	opt = xchg(&np->opt, opt);
	sk_dst_reset(sk);

out:
	if (opt)
		sock_kfree_s(sk, opt, opt->tot_len);

	return retval;
}

/*
 * Does the packet have other HBH options besides PLA?
 * Return: true/false.
 */
static int find_hbh_not_pla(struct ipv6_hopopt_hdr *hbh)
{
	int hbh_len = ipv6_optlen(hbh);
	int offset = 2;

	char *raw = (char *) hbh;

	while(offset < hbh_len)
		switch (raw[offset]) {
		case IPV6_TLV_PAD0:
			offset++;
			break;

		case IPV6_TLV_PADN:
		case OPTION_PLA:
			offset += raw[offset + 1] + 2;
			break;
			
		default:
			return 1;
			break;
		}

	/* Nothing found */
	return 0;
}


signed int pla_strip_pla_header(struct sk_buff **skb)
{
	struct ipv6hdr *ip6hdr;
	struct ipv6_hopopt_hdr *hopopt;
	int hopopt_len;
	int err = 0;

	/* We need a new copy of the packet if someone else uses it */
	if (skb_cloned(*skb)) {
		struct sk_buff *nskb;

		if ( (nskb = skb_copy(*skb, GFP_ATOMIC)) == NULL) {
			err = -ENOMEM;
			goto drop_out;
		}
       
		kfree_skb(*skb);
		*skb = nskb;
	}

	ip6hdr = (*skb)->nh.ipv6h;
	hopopt = (struct ipv6_hopopt_hdr*) (ip6hdr + 1);
	hopopt_len = ipv6_optlen(hopopt);

	// jlu XXX: If we have multiple header we should remove the last one... TBD */
	/* Is PLA the only option in the HBH header? */
	if (find_hbh_not_pla(hopopt)) {
		/* Remove only the PLA option */
		/* Currently, this only replaces the PLA option with padding. */
		/* jlu XXX: Untested code. */
		struct ipv6_opt_hdr *opt_hdr = (struct ipv6_opt_hdr*) ( ((char*) pla_find_pla_header(*skb)) - 2);

		memset(opt_hdr + 1, 0, opt_hdr->hdrlen);
		opt_hdr->nexthdr = IPV6_TLV_PADN;

	} else {
		/* Remove the entire HBH header */
		ip6hdr->nexthdr = hopopt->nexthdr;
		ip6hdr->payload_len = htons(ntohs(ip6hdr->payload_len) - hopopt_len);
		
		memmove( ((char*) ip6hdr) + hopopt_len, ip6hdr, sizeof(struct ipv6hdr) );
		(*skb)->nh.raw += hopopt_len;
		
		skb_pull(*skb, hopopt_len);
	}

	return 0;

drop_out:
	return err;
}

int pla_header_new_pla_hbh(struct ipv6hdr *ipv6h, char *res_buffer)
{
	struct ipv6_hopopt_hdr *hop_hdr = (struct ipv6_hopopt_hdr*) res_buffer;

	memcpy(res_buffer, &hbh_and_pla, sizeof(hbh_and_pla));

	hop_hdr->nexthdr = ipv6h->nexthdr;

	return sizeof(hbh_and_pla);
}

/* 
 * Add PLA to existing HBH header. 
 * Returns: The length increase of the HBH header (always a multiple of 8).
 */
int pla_header_add_pla_hbh(struct ipv6hdr *ipv6h, char *res_buffer)
{
	struct ipv6_hopopt_hdr *old_hbh_hdr = (struct ipv6_hopopt_hdr*) (ipv6h + 1);
	struct ipv6_hopopt_hdr *new_hbh_hdr = (struct ipv6_hopopt_hdr*) res_buffer;

	char *cur_ptr = res_buffer;
	char *new_option_start;

	memcpy(cur_ptr, old_hbh_hdr, ipv6_optlen(old_hbh_hdr));
	cur_ptr += ipv6_optlen(old_hbh_hdr);

	new_option_start = cur_ptr;

	/* PLA is aligned 8n+4. We need 2 bytes of padding before PLA TLV start */
	cur_ptr[0] = IPV6_TLV_PADN;
	cur_ptr[1] = 2 - 2;
	cur_ptr += 2;

	/* PLA_TLV */
	cur_ptr[0] = OPTION_PLA;
	cur_ptr[1] = sizeof(struct pla_hdr);
	cur_ptr += sizeof(struct ipv6_opt_hdr);

	/* Data itself is added later */
	memset(cur_ptr, 0, sizeof(struct pla_hdr));
	cur_ptr += sizeof(struct pla_hdr);

	/* Pad 4 bytes to complete 8 byte alignment */
	cur_ptr[0] = IPV6_TLV_PADN;
	cur_ptr[1] = 4 - 2;
	cur_ptr += 4;

	new_hbh_hdr->hdrlen = ((cur_ptr - res_buffer) >> 3) - 1;

	return cur_ptr - new_option_start;
}


/*
 * Copies/creates the complete new HBH option into res_buffer.
 * Returns the length increase of the returned header (in bytes)
 */
int pla_header_modify_hbh_add_pla(struct ipv6hdr *ipv6h, char *res_buffer)
{
	int add_len;

	if (ipv6h->nexthdr == NEXTHDR_HOP) 
		add_len = pla_header_add_pla_hbh(ipv6h, res_buffer);
	else
		add_len = pla_header_new_pla_hbh(ipv6h, res_buffer);

	return add_len;
}

int pla_add_header(struct sk_buff **skb)
{
	struct ipv6hdr *ipv6h;
	struct sk_buff *newskb;

	char hbh_result[256];  // jlu XXX: A fixed size array...
	int hbh_len_add;

	struct ipv6_hopopt_hdr *hop_hdr = (struct ipv6_hopopt_hdr*) &hbh_result[0];
	
	hbh_len_add = pla_header_modify_hbh_add_pla( (*skb)->nh.ipv6h, &hbh_result[0]);

	/* skb->dst may not be set on packets meant for link local delivery */
        if ((*skb)->dst && (*skb)->len + hbh_len_add > dst_mtu((*skb)->dst)) {
                /* Again, force OUTPUT device used as source address */
                (*skb)->dev = (*skb)->dst->dev;
                icmpv6_send(*skb, ICMPV6_PKT_TOOBIG, 0, dst_mtu((*skb)->dst) - hbh_len_add, (*skb)->dev);
                return -EMSGSIZE;
        }

	/* Allocate a big enough skb */
	newskb = skb_copy_expand(*skb, skb_headroom(*skb) + hbh_len_add, skb_tailroom(*skb), GFP_ATOMIC);
	if (!newskb)
		return -ENOMEM;
	
	if ((*skb)->sk)
		skb_set_owner_w(newskb, (*skb)->sk);
	kfree_skb(*skb);
	*skb = newskb;
	
	if (!skb_make_writable(skb, (*skb)->len))
		return -ENOMEM;

	skb_push(*skb, hbh_len_add);
	memmove((*skb)->nh.raw - hbh_len_add, (*skb)->nh.raw, sizeof(struct ipv6hdr) );
	
	/* Update pointers broken by the move. */
	(*skb)->nh.raw -= hbh_len_add;
	ipv6h = (*skb)->nh.ipv6h;

	/* Add the hop-by-hop option + PLA. */
	memcpy(ipv6h + 1, hop_hdr, ipv6_optlen(hop_hdr) );
	
	ipv6h->nexthdr = NEXTHDR_HOP;
	ipv6h->payload_len = htons(ntohs(ipv6h->payload_len) + hbh_len_add);

	return 0;
}
