#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/kernel.h>
#include <linux/timer.h>
#include <linux/string.h>
#include <linux/sockios.h>
#include <linux/net.h>
#include <linux/fcntl.h>
#include <linux/skbuff.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <net/sock.h>
#include <linux/init.h>
#include <linux/spinlock.h>

#include <linux/netlink.h>

#include "pla_crypto.h"
#include "pla_netlink.h"
#include "pla_policy.h"
#include "pla_src_id.h"

struct sock *planl = NULL;

DECLARE_MUTEX(netlink_user_mutex);     /* Mutex for threads entering netlink from user space */

static int pla_netlink_rcv_msg(struct sk_buff *skb, struct nlmsghdr *nlh, int *errp)
{
	switch(nlh->nlmsg_type) {
	case  NLMSG_PLA_ADD:
		pla_crypto_netlink(nlh);
		break;

	case NLMSG_PLA_SET_INTERFACES:
		pla_policy_netlink(nlh);
		break;

	case NLMSG_PLA_SET_SRC:
		pla_src_id_netlink(nlh);
		break;

	default:
		break;
	}

	*errp = 0;

	return 0;
}


static int pla_netlink_rcv_skb(struct sk_buff *skb)
{
        int err;
        struct nlmsghdr *nlh;

        while (skb->len >= NLMSG_SPACE(0)) {
                u32 rlen;

                nlh = (struct nlmsghdr *)skb->data;
                if (nlh->nlmsg_len < sizeof(struct nlmsghdr) || skb->len < nlh->nlmsg_len)
                        return 0;

                rlen = NLMSG_ALIGN(nlh->nlmsg_len);
                if (rlen > skb->len)
                        rlen = skb->len;

                if (pla_netlink_rcv_msg(skb, nlh, &err)) {
                        if (!err)
                                return -1;
                        netlink_ack(skb, nlh, err);
                } else
                        if (nlh->nlmsg_flags & NLM_F_ACK)
                                netlink_ack(skb, nlh, 0);
                skb_pull(skb, rlen);
        }

        return 0;
}

static void pla_netlink_rcv(struct sock *sk, int len)
{
	struct sk_buff *skb;

	if (down_interruptible(&netlink_user_mutex))
		goto out;

	while ((skb = skb_dequeue(&sk->sk_receive_queue)) != NULL) {
		pla_netlink_rcv_skb(skb);
		kfree_skb(skb);
	}

	up(&netlink_user_mutex);

out:
	return;
}

void pla_netlink_uninit(void)
{
        sock_release(planl->sk_socket);
	planl = NULL;
        return;
}

int pla_netlink_init(void)
{
        planl = netlink_kernel_create(NETLINK_PLA, PLAGRP_MAX, pla_netlink_rcv, THIS_MODULE);
        if (!planl) {
                printk(KERN_ERR "cannot initialize PLA netlink!\n");
                return -1;
        }

        return 0;
}
