#include <linux/module.h>
#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/list.h>

#include <linux/ip.h>     /* Needed here. Otherwise the linksys target fails! */

#include "pla_state.h"

spinlock_t state_lock;
struct list_head state_list;

/*
 *  Copy TTP public key to ttpkey.
 *  Returns: 0 on success;
 *           -1: unknown TTP and request sent recently.
 *           -2: unknown TTP, no request sent recently.  jlu XXX: TBD
 */
int pla_state_get(char *ttpkey)
{
	// jlu XXX: Implement the ICMP check.

	struct list_head *cursor;

	spin_lock_bh(&state_lock);

	list_for_each(cursor, &state_list) {
		struct sstate *this_entry = (struct sstate*) cursor;

		memcpy(ttpkey, this_entry->ttp_pubkey, this_entry->ttp_pubkey_len);
		spin_unlock_bh(&state_lock);
		return 0;
	}

	spin_unlock_bh(&state_lock);
	return -1;
}

/*
 * Add new TTP key
 * Returns 0 on success.
 * negative on error.
 */
int pla_state_add(char *ttpkey, int ttpkey_len)
{
	int retval = 0;
	struct sstate *sentry = NULL;
	char *state_mem = NULL;

	state_mem = kmalloc(ttpkey_len, GFP_ATOMIC);
	sentry = kmalloc(sizeof(struct sstate), GFP_ATOMIC);
	if (!state_mem || !sentry) {
		if (state_mem)
			kfree(state_mem);
		if (sentry)
			kfree(sentry);
		retval = -ENOMEM;
		goto out_err;
	}
	sentry->state_mem = state_mem;

	sentry->ttpid = 1;         // jlu XXX: Should be set from parameter.

	sentry->ttp_pubkey_len = ttpkey_len;
	sentry->ttp_pubkey = (char*) state_mem;
	memcpy(sentry->ttp_pubkey, ttpkey, sentry->ttp_pubkey_len);

	spin_lock_bh(&state_lock);
	list_add(&sentry->list_head, &state_list);
	spin_unlock_bh(&state_lock);

out_err:
	return retval;
};

void pla_state_flush(void)
{
	struct list_head *cursor;
	struct list_head *next;

	/* jlu XXX: Should we take the lock before deleting? */
	list_for_each_safe(cursor, next, &state_list) {
		struct sstate *this_entry = (struct sstate*) cursor;

		kfree(this_entry->state_mem);

		list_del(cursor);
	}

	return;
}


int pla_state_init(void)
{
	spin_lock_init(&state_lock);
	
	INIT_LIST_HEAD(&state_list);

	return 0;
}

void pla_state_uninit(void)
{
	pla_state_flush();
       
	return;
}
