#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <unistd.h>

#include <linux/types.h>

#include "ec.h"
#include "ec_jlu.h"


/*
 * Binary certificate to parameters
 * Returns: length of consumed data
 */
int ec_bin_to_cert(char *source, elem_t *rho, int8_t *b)
{
	char *cur = source;

	/* import rho */
	memcpy(rho, cur, sizeof(elem_t));
	cur += sizeof(elem_t);
	
	/* import b */
	memcpy(b, cur, sizeof(*b));
	cur += sizeof(*b);

	return cur - source;
}

/*
 * Returns the length of the created certificate.
 */
int ec_cert_to_bin(char *target, elem_t rho, int8_t b)
{
	char *cur = target;

	/* export rho */
	memcpy(cur, rho, sizeof(elem_t));
	cur += sizeof(elem_t);
	
	/* export b */
	memcpy(cur, &b, sizeof(b));
	cur += sizeof(b);

	return cur - target;
}


/*
 * Returns length of consumed data.
 */
int ec_bin_to_privkey(char *source, mpz_t *sigma)
{
	char *cur = source;
	__u8 count;

	/* import sigma. Client private key */
	memcpy(&count, cur, sizeof(count));
	cur += sizeof(count);

	mpz_import(*sigma, count, 1, sizeof(__u32), 1, 0, cur);
//	mpz_out_str(stderr, 10, *sigma); fprintf(stderr, "\n");
	cur += sizeof(__u32) * count;

	return cur - source;
}

/* 
 */
int ec_bin_to_ttppubkey(char *source)
{
	char *cur = source;
	
	/* import e.y. (TTP public key) */
	memcpy(&e.y, cur, sizeof(e.y));
	cur += sizeof(e.y);
	
	return cur - source;
}

int ec_privkey_to_bin(char *target, mpz_t sigma)
{
	char *cur = target;
	size_t countp;

	/* export sigma (private key) */
	mpz_out_str(stderr, 10, sigma);

	__u32 tmpbuf[100];
	mpz_export(&tmpbuf, &countp, 1, sizeof(tmpbuf[0]), 1, 0, sigma);
	
	*cur = countp;
	cur += 1;
	
	memcpy(cur, tmpbuf, sizeof(tmpbuf[0]) * countp);
	cur += sizeof(tmpbuf[0]) * countp;

	return cur - target;
}


/*
 * Convert certificate variables to binary.
 * Returns length of result.
 */
int ec_params_to_bin(char *target, elem_t rho, mpz_t sigma, int b)
{
	char *cur = target;

	/* Export certificate */
	cur += ec_cert_to_bin(cur, rho, b);

	/* Export private key */
	cur += ec_privkey_to_bin(cur, sigma);

	/* export e.y (TTP public key) */
	memcpy(cur, &e.y, sizeof(e.y));
	cur += sizeof(e.y);

	return cur - target;
}

int ec_sig_to_bin(char *target, ecselfsig_t *sig)
{
	char *cur = target;

	memcpy(cur, sig->rho, sizeof(sig->rho));
	cur += sizeof(sig->rho);

	memcpy(cur, sig->r, sizeof(sig->r));
	cur += sizeof(sig->r);

	memcpy(cur, &sig->b, sizeof(sig->b));
	cur += sizeof(sig->b);

	/* Finally, the mpz... grr...*/
	__u32 tmpbuf[100];
	size_t countp;
	mpz_export(&tmpbuf, &countp, 1, sizeof(tmpbuf[0]), 1, 0, sig->s);
	*cur = countp;
	cur += 1;
	memcpy(cur, tmpbuf, sizeof(tmpbuf[0]) * countp);
	cur += sizeof(tmpbuf[0]) * countp;

	return cur - target;
}


int ec_bin_to_sig(char *source, ecselfsig_t *sig)
{
	char *cur = source;

	memcpy(sig->rho, cur, sizeof(sig->rho));
	cur += sizeof(sig->rho);

	memcpy(sig->r, cur, sizeof(sig->r));
	cur += sizeof(sig->r);

	memcpy(&sig->b, cur, sizeof(sig->b));
	cur += sizeof(sig->b);

	/* Finally, the mpz... grr...*/
	__u32 tmpbuf[100];
	size_t countp;

	countp = *cur;
	cur += 1;

	memcpy(tmpbuf, cur, sizeof(tmpbuf[0]) * countp);

	mpz_init(sig->s);
	mpz_import(sig->s, countp, 1, sizeof(__u32), 1, 0, tmpbuf);

	cur += sizeof(tmpbuf[0]) * countp;

	return cur - source;
}
