#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <netinet/in.h>
#include <openssl/rsa.h>
#include <openssl/objects.h>

#include "usercrypto.h"
#include "../src/configs.h"
#include "ec.h"
#include "ec_jlu.h"

static void hexdump(unsigned char *buf, unsigned int len)
{
        while (len--)
                fprintf(stderr, "%02x", *buf++);
}

void rsa_request(uint32_t idnum, int type, char *opdata) 
{
	struct crypto_sign_request *sr;
	struct crypto_verify_request *vr;

	char resultdata[512];   /* jlu XXX: Big enough buffer */
	int resultlen = 0;

	struct crypto_sign_response *sres;
	struct crypto_verify_response *vres;
	ecselfsig_t sig;

//	fprintf(stderr, "IDnum: %d\n", idnum);

	switch(type) {
	case CRYPTO_TASK_TYPE_SIGN:
	{
		int8_t b;
		mpz_t sigma;

		mpz_init(sigma);

		sr = (struct crypto_sign_request*) opdata;
		sres = (struct crypto_sign_response*) &resultdata[0];

//		fprintf(stderr, "S");

		ec_bin_to_privkey(sr->privkey, &sigma);
		ec_bin_to_cert(sr->cert, &sig.rho, &b);
		sig.b = b;

		ec_sign_self(&sig, sr->hash, sigma);
		ec_sig_to_bin(sres->signature, &sig);

		mpz_clear(sigma);

		resultlen = PLA_SIG_LEN;
		break;
	}
	case CRYPTO_TASK_TYPE_VERIFY:
	{
		/* Workaround for a bug in EC code. EC expects a NULL terminated id_hash. */
		char id_hash_buf[PLA_HASH_DIGEST_LENGTH + 1];
		id_hash_buf[PLA_HASH_DIGEST_LENGTH] = 0;

		vr = (struct crypto_verify_request*) opdata;
		vres = (struct crypto_verify_response*) &resultdata[0];

//		fprintf(stderr, "V");

		/* Copy TTP public key to place */
		memcpy(&e.y, vr->ttp_pubkey, sizeof(e.y));
		memcpy(id_hash_buf, vr->id_hash, PLA_HASH_DIGEST_LENGTH);

//		memset(&sig, 0, sizeof(sig));   // jlu XXX: Is this needed?
		ec_bin_to_sig(vr->signature, &sig);

		vres->result = ec_verify_self(&sig, vr->hash, id_hash_buf);
//		fprintf(stderr, "%s\n", id_hash_buf);
		resultlen = 1;

		break;
	}
	default:
		assert(0);
		break;
	}

	usercrypto_result(idnum, &resultdata[0], resultlen);
	return;
}

void null_request(uint32_t idnum, int type, char *opdata) 
{
	struct crypto_sign_request *sr;
	struct crypto_verify_request *vr;

	char resultdata[100]; 
	int resultlen = 0;

	struct crypto_sign_response *sres;
	struct crypto_verify_response *vres;
	ecselfsig_t sig;

	switch(type) {
	case CRYPTO_TASK_TYPE_SIGN:
		resultlen = PLA_SIG_LEN;
		break;

	case CRYPTO_TASK_TYPE_VERIFY:
		vres = (struct crypto_verify_response*) &resultdata[0];
		vres->result = 1;
		resultlen = 1;
		break;

	default:
		assert(0);
		break;
	}

	usercrypto_result(idnum, &resultdata[0], resultlen);
	return;
}


struct crypto_ops rsa_ops =
{
	.request = rsa_request
};

struct crypto_ops null_ops = 
{
	.request = null_request
};

int main(int argc, char **argv)
{
	int null_crypto = 0;
	int c;

	while (1) {
		c = getopt_long (argc, argv, "n", NULL, NULL);
		if (c == -1)
			break;
		
		switch (c) {
		case 'n':
			null_crypto = 1;
			break;

		case '?':
			exit(EXIT_FAILURE);
			
		default:
			fprintf (stderr, "?? getopt returned character code 0%o ??\n", c);
			exit(EXIT_FAILURE);

		}
	}
	

	if (null_crypto) {
		usercrypto_run(&null_ops);
	} else {
		ec_init();
		usercrypto_run(&rsa_ops);
	}

	return EXIT_SUCCESS;
}
