#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>

#include <linux/socket.h> 
#include <linux/types.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <linux/netlink.h>

#include <net/if.h>

#include "../src/pla_netlink.h"
#include "../src/configs.h"

#include "ec.h"

#define MAX_PAYLOAD 1024  /* maximum payload size */

#define KEYCONF_PORT 0x1234

static int netlink_fd;

void nlh_fill(struct nlmsghdr* nlh, int type, int len, int flags)
{
	nlh->nlmsg_len = NLMSG_SPACE(len);
	nlh->nlmsg_pid = getpid();
	nlh->nlmsg_type = type;
	nlh->nlmsg_flags = flags;

	return;
}

int netlink_socket(void)
{
	int sock_fd;
	struct sockaddr_nl src_addr;
		
	sock_fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_PLA);
	if (sock_fd < 0) {
		perror("socket");
		exit(EXIT_FAILURE);
	}

	memset(&src_addr, 0, sizeof(struct sockaddr_nl));
	src_addr.nl_family = AF_NETLINK;      
	src_addr.nl_pid = getpid(); 
	src_addr.nl_groups = 0;

	if (bind(sock_fd, (struct sockaddr*) &src_addr, sizeof(src_addr)) < 0) {
		perror("bind");
		exit(EXIT_FAILURE);
	}
	
	return sock_fd;
}

int send_netlink(void *packet, int len, int type)
{
	struct sockaddr_nl dest_addr;

	struct nlmsghdr *nlh = NULL;
	struct iovec iov;
	struct msghdr msg;

	memset(&dest_addr, 0, sizeof(dest_addr));
	dest_addr.nl_family = AF_NETLINK;
	dest_addr.nl_pid = 0;  
	dest_addr.nl_groups = 0;

	nlh = malloc(NLMSG_SPACE(MAX_PAYLOAD));
	if (!nlh) {
		fprintf(stderr, "Out of memory in placonf.\n");
		exit(EXIT_FAILURE);
	}

	nlh_fill(nlh, type, MAX_PAYLOAD, 0);

	memcpy(NLMSG_DATA(nlh), packet, len);

	if (sendto(netlink_fd, nlh, nlh->nlmsg_len, 0, (struct sockaddr*) &dest_addr, sizeof(dest_addr)) < 0) {
		perror("sendto");
		exit(EXIT_FAILURE);
	}

	return 0;
}

void set_userspace_target(char *target)
{
	struct pla_netlink_target userspace_target;

	memset(&userspace_target, 0, sizeof(userspace_target));

	if (inet_pton(AF_INET, target, &userspace_target.addr) <= 0) {
		fprintf(stderr, "Error in address parameter\n");
		exit(EXIT_FAILURE);
	}

	send_netlink(&userspace_target, sizeof(userspace_target), NLMSG_PLA_ADD );

	return;
}

int set_interfaces(int argc, char **argv)
{
	int if_in, if_out;
	char* policystr;
	int policynum;

	if (argc != 3) {
		fprintf(stderr, "Invalid number of parameters\n");
		exit(EXIT_FAILURE);
	}

	if_in = if_nametoindex(argv[optind]);
	if (if_in == 0) {
		fprintf(stderr, "No such interface: %s\n", argv[optind]);
		exit(EXIT_FAILURE);
	}
	optind++;

	if_out = if_nametoindex(argv[optind]);
	if (if_out == 0) {
		fprintf(stderr, "No such interface: %s\n", argv[optind]);
		exit(EXIT_FAILURE);
	}
	optind++;

	policystr = argv[optind];
	if (strcmp(policystr, "drop") == 0)
		policynum = PLA_FORWARD_POLICY_DROP;
	else if (strcmp(policystr, "add") == 0)
		policynum = PLA_FORWARD_POLICY_SIGN;
	else if (strcmp(policystr, "verify") == 0)
		policynum = PLA_FORWARD_POLICY_VERIFY;
	else if (strcmp(policystr, "strip") == 0)
		policynum = PLA_FORWARD_POLICY_STRIP;
	else {
		fprintf(stderr, "Invalid policy string: %s\n",  policystr);
		exit(EXIT_FAILURE);
	}

	if (if_out == 0 && strcmp(argv[optind], "local") != 0) {
		fprintf(stderr, "No such interface: %s\n", argv[optind]);
		exit(EXIT_FAILURE);
	}
	optind++;

	fprintf(stderr, "Setting policy: %d\n", policynum);

	struct pla_netlink_set_interface if_set = {
		.if_in  = if_in,
		.if_out = if_out,
		.policy = policynum,
		};

	
	fprintf(stderr, "Indexes: %d %d\n", if_set.if_in, if_set.if_out);

	send_netlink(&if_set, sizeof(if_set), NLMSG_PLA_SET_INTERFACES );

	return 0;
}

void daemon_mode(void)
{
	socklen_t fromlen;
	struct sockaddr_in sin;
	int s;
	char inbuf[1500];
	
	mpz_t x; /* domain private key of TTP */

	ec_init();

	mpz_init(x);
	ec_keygen(&(e.y),x); /* e.y is domain public key of TTP */

	if ( (s = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(EXIT_FAILURE);
	}

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(KEYCONF_PORT);

	if (bind(s, (struct sockaddr*) &sin, sizeof(sin)) < 0) {
		perror("bind");
		exit(EXIT_FAILURE);
	}

	while(1) {
		mpz_t k;
		mpz_t rhoi;
		mpz_t sigma;
		char target[512];
		ecpoint_t kgd;
		int len;

		fromlen = sizeof(sin);
		memset(inbuf, 0, sizeof(inbuf));
		len = recvfrom(s, inbuf, sizeof(inbuf), 0, (struct sockaddr*) &sin, &fromlen);
		if (len < 0) {
			perror("recvfrom");
			exit(EXIT_FAILURE);
		}

		fprintf(stderr, "ID hash is: %s\n", inbuf);

		/* Create new key for the client */
		mpz_init(k);
		ec_keygen(&kgd,k);

		int b;
		ec_compress_point(&b,&kgd);
		elem_t rho, hid;
		elem_import_bytes(hid,inbuf); 
		elem_chomp(hid); /* hash of the ID */
		gnb_add(rho,hid,kgd.x);
		
		mpz_init(rhoi);
		mpz_init(sigma);
		elem_to_int(rhoi,rho);
		mpz_fdiv_r(rhoi,rhoi,e.r);
		mpz_mul(sigma,rhoi,x);
		mpz_add(sigma,k,sigma);
		mpz_fdiv_r(sigma,sigma,e.r);

		/* Create response from rho, sigma, b, (e.y?) */
		len = ec_params_to_bin(&target[0], rho, sigma, b);
		
		if (sendto(s, target, len, 0, (struct sockaddr*) &sin, sizeof(sin)) < 0) {
			perror("sendto");
			exit(EXIT_FAILURE);
		}
		
		fprintf(stderr, "Sent %d bytes\n", len);

		mpz_clear(sigma);
		mpz_clear(rhoi);
		mpz_clear(k);
	}


	close(s);

	return;
}

int run_ec_tests(elem_t *rho, int8_t b, mpz_t sigma)
{
	int result;

	ecselfsig_t sig;
	char message_hash[] = "98765432109876543210";
	char id_hash[] = "01234567890123456789";

	ec_init();

	memset(&sig, 0, sizeof(sig));
	memcpy(&sig.rho, rho, sizeof(*rho));
	sig.b = b;
	ec_sign_self(&sig, message_hash, sigma);
	
	result = ec_verify_self(&sig, message_hash, id_hash);
	switch(result) {
	case 1:
		fprintf(stderr, "ECC tests successful\n");
		break;
	default:
		fprintf(stderr, "ECC tests failed!!!!\n");
		exit(EXIT_FAILURE);
		break;
	}

	mpz_clear(sig.s);
	return 0;
}

void client_mode(char *address)
{
	char id_hash[PLA_HASH_DIGEST_LENGTH];
	struct sockaddr_in sin;
	int s, i;
	int len;
	char inbuf[1500];

	fprintf(stderr, "Address is: %s\n", address);
	
	if ( (s = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(EXIT_FAILURE);
	}

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(KEYCONF_PORT);

	if (bind(s, (struct sockaddr*) &sin, sizeof(sin)) < 0) {
		perror("bind");
		exit(EXIT_FAILURE);
	}

	if (inet_pton(AF_INET, address, &sin.sin_addr) <= 0) {
		fprintf(stderr, "Invalid address\n");
		exit(EXIT_FAILURE);
	}

#if 1
	/* jlu XXX: Yes, this does it in the dumbest possible way. */
	fprintf(stderr, "Creating hash ID\n");
	srandom(time(NULL));
	for (i=0; i<PLA_HASH_DIGEST_LENGTH; i++)
		id_hash[i] = random();
#else
	strncpy(id_hash, "01234567890123456789", PLA_HASH_DIGEST_LENGTH);
#endif

	fprintf(stderr, "Sending request\n");
	if (sendto(s, id_hash, PLA_HASH_DIGEST_LENGTH, 0, (struct sockaddr*) &sin, sizeof(sin)) < 0) {
		perror("sendto");
		exit(EXIT_FAILURE);
	}

	len = recv(s, &inbuf, sizeof(inbuf), 0);
	fprintf(stderr, "Received response of %d bytes\n", len);

	close(s);

	int8_t b;
	elem_t rho;

	mpz_t sigma;
	mpz_init(sigma);

	int privkeylen;
	int certlen;

	/* Decode response packet */
	char *cur = inbuf;
	certlen = ec_bin_to_cert(cur, &rho, &b);
	cur += certlen;

	privkeylen = ec_bin_to_privkey(cur, &sigma);
	cur += privkeylen;

	cur += ec_bin_to_ttppubkey(cur);

#if 0
	run_ec_tests(&rho, b, sigma);
#endif
        
        /* Send certificate and private key to kernel over netlink */
	struct pla_netlink_add_src *src;
	char *tmp;

	int pubkey_len = sizeof(e.y);

	int src_len = sizeof(struct pla_netlink_add_src) + pubkey_len + PLA_PRIVKEY_LEN + PLA_CERT_LEN + PLA_HASH_DIGEST_LENGTH;
	src = malloc(src_len);
	if (!src) {
		fprintf(stderr, "Out of memory in client_mode\n");
		exit(EXIT_FAILURE);
	}

	src->publen = pubkey_len;
	src->privlen = PLA_PRIVKEY_LEN;
	src->certlen = PLA_CERT_LEN;
	src->id_hashlen = PLA_HASH_DIGEST_LENGTH;

	tmp = &src->data[0];

	/* TTP public key */
	memcpy(tmp, &e.y, pubkey_len);
	tmp += src->publen;
	
	memset(tmp, 0xAB, src->privlen);  // jlu XXX: deadbeef
	ec_privkey_to_bin(tmp, sigma);
	tmp += PLA_PRIVKEY_LEN;
	
	memset(tmp, 0xCC, src->certlen);  // jlu XXX: deadbeef
	int clen = ec_cert_to_bin(tmp, rho, b);
	tmp += PLA_CERT_LEN;
	fprintf(stderr, "Length of cert: %d\n", clen);

	memcpy(tmp, id_hash, src->id_hashlen);
	tmp += src->id_hashlen;
	
	send_netlink(src, src_len, NLMSG_PLA_SET_SRC);

	free(src);

	return;
}

int main(int argc, char **argv) 
{
	int c;

	netlink_fd = netlink_socket();
	if (netlink_fd < 0) {
		perror("placonf");
		exit(EXIT_FAILURE);
	}

	while (1) {
		c = getopt_long (argc, argv, "a:idc:", NULL, NULL);
		if (c == -1)
			break;
		
		switch (c) {
		case 'a':
			set_userspace_target(optarg);
			break;
			
		case 'i':
			set_interfaces(argc - 2, argv);
			break;

		case 'd':
			daemon_mode();
			break;

			/* jlu XXX: Client mode needs to merged into set_source_id... */
		case 'c':
			client_mode(optarg);
			break;

		case '?':
			exit(EXIT_FAILURE);
			
		default:
			fprintf (stderr, "?? getopt returned character code 0%o ??\n", c);
			exit(EXIT_FAILURE);

		}
	}
	
	if (optind < argc) {
		fprintf (stderr, "non-option ARGV-elements: ");
		while (optind < argc)
			fprintf (stderr, "%s ", argv[optind++]);
		printf ("\n");
	}


	/* Close Netlink Socket */
	close(netlink_fd);

	return 0;
}    
